package org.expeditee.core;

import org.expeditee.core.bounds.AxisAlignedBoxBounds;

/**
 * An area to clip drawing operations to.
 * 
 * @author cts16
 */
public class Clip {
	
	/** The bounds of this clip. Null means there is no clipping. */
	private AxisAlignedBoxBounds _bounds = null;
	
	/** Flag to identify when there is no unclipped area. */
	private boolean _isFullyClipped = false;
	
	/** Default constructor. */
	public Clip()
	{
	}
	
	/** Construct a clip encompassing the given bounds. */
	public Clip(AxisAlignedBoxBounds bounds)
	{
		this(bounds, false);
	}
	
	/** Standard constructor. */
	protected Clip(AxisAlignedBoxBounds bounds, boolean isFullyClipped)
	{
		if (bounds != null) _bounds = bounds.clone();
		_isFullyClipped = isFullyClipped;
	}
	
	/** Copy constructor. */
	public Clip(Clip other)
	{
		if (other._bounds != null) _bounds = other._bounds.clone();
		_isFullyClipped = other._isFullyClipped;
	}

	/** Creates a clip which clips the entire area. */
	public static Clip getFullClip()
	{
		return new Clip(null, true);
	}
	
	/** Checks if this clip clips all possible space. */
	public boolean isFullyClipped()
	{
		return _isFullyClipped;
	}
	
	/** Checks if this clip is not clipping anything. */
	public boolean isNotClipped()
	{
		return !isFullyClipped() && _bounds == null;
	}
	
	/**
	 * Gets the bounds of the clip. Null means no clipping.
	 * Should check isFullyClipped() to see if the entire area is clipped.
	 */
	public AxisAlignedBoxBounds getBounds()
	{
		if (_bounds == null) return null;
		
		return _bounds.clone();
	}
	
	/** Resets the clip so that no area is clipped. */
	public Clip clear()
	{
		_bounds = null;
		_isFullyClipped = false;
		
		return this;
	}
	
	/** Expands the clip area to incorporate the given bounds. */
	public Clip includeBounds(AxisAlignedBoxBounds bounds)
	{
		if (bounds == null) return this;
		
		// If we are fully clipped then the given bounds will be the only unclipped area
		if (isFullyClipped()) {
			_isFullyClipped = false;
			_bounds = bounds.clone();
			
		// If _bounds is null then we already contain the given bounds
		} else if (_bounds != null) {
			_bounds = _bounds.combineWith(bounds);
		}
		
		return this;
	}
	
	/** Updates the clip to include only those areas that are inside the given bounds. */
	public Clip intersectWith(AxisAlignedBoxBounds bounds)
	{
		if (bounds == null) return this;
		
		// If fully clipped then no area is inside the given bounds
		if (isFullyClipped()) return this;
		
		// If there is no clipped area then the given bounds defines the intersection
		if (_bounds == null) {
			_bounds = bounds.clone();
		} else {
			AxisAlignedBoxBounds newClip = _bounds.intersectionWith(bounds);
			
			// If the current bounds and given bounds don't intersect, the area is fully clipped
			if (newClip == null) {
				_isFullyClipped = true;
			} else {
				_bounds = newClip;
			}
		}
		
		return this;
	}
	
	/** Updates the clip to include only those areas that are inside the given clip. */
	public Clip intersectWith(Clip clip)
	{
		if (clip == null) return this;
		
		if (!clip.isFullyClipped()) return intersectWith(clip.getBounds());
		
		_isFullyClipped = true;
		
		return this;
	}
	
	@Override
	public Clip clone()
	{
		return new Clip(this);
	}
}
