package org.expeditee.core;

import java.util.Stack;

public class EnforcedClipStack {

	/** The stack of sub-area clips. */
	private Stack<Clip> _clipStack;
	
	/** Default constructor. */
	public EnforcedClipStack()
	{
		_clipStack = new Stack<Clip>();
	}
	
	/** Pushes a new clip onto the stack, ensuring that it is a sub-area of the top of the stack. */
	public EnforcedClipKey push(Clip clip)
	{
		// No null clips allowed
		if (clip == null) return null;
		
		// Leave the given clip unchanged (in case the provider wants to keep using it)
		clip = clip.clone();
		
		// Each clip must be a sub-area of the one below it in the stack
		if (!_clipStack.isEmpty()) {
			clip.intersectWith(_clipStack.peek());
		}
		
		// Put the new clip on the stack
		_clipStack.push(clip);
		
		// Return the clipped area
		return EnforcedClipKey.get(clip);
	}
	
	/** Pops the top of the stack if there given key corresponds to it. */
	public Clip pop(EnforcedClipKey key)
	{
		if (!_clipStack.isEmpty() && key != null && key.matches(_clipStack.peek())) return _clipStack.pop();
		
		return null;
	}
	
	/** Returns a copy of the current top of the stack. */
	public Clip peek()
	{
		if (!_clipStack.isEmpty()) return _clipStack.peek().clone();
		
		return null;
	}
	
	/** Key class to ensure only the owner of a given pushed clip can pop it again. */
	public static class EnforcedClipKey {
		
		/** The clip this key corresponds to. */
		private Clip _keyedClip;
		
		/** Gets a key for the given clip. */
		public static EnforcedClipKey get(Clip clip)
		{
			if (clip == null) return null;
			
			return new EnforcedClipKey(clip);
		}
		
		/** Private constructor so clip cannot be null. */
		private EnforcedClipKey(Clip clip)
		{
			_keyedClip = clip;
		}
		
		/** Whether this key corresponds to the given clip. */
		public boolean matches(Clip clip)
		{
			return clip == _keyedClip;
		}
		
	}
	
}
