package org.expeditee.core;

/**
 * Handle-type for fonts managed by the font manager. Should be able to
 * instantiate your own font objects, and have the font manager use an appropriate
 * internal font when necessary (i.e. Font objects don't need to be constructed by
 * the font manager, they are interpreted on use).
 * 
 * @author cts16
 */
public class Font {

	/** Enumeration of the possible styles of font (bold/italic). */
	public enum Style {
		PLAIN,
		BOLD,
		ITALIC,
		BOLD_ITALIC
	}

	/** The default family for fonts which don't explicitly specify one. */
	public static final String DEFAULT_FAMILY_NAME = "Serif";
	/** The default style for fonts which don't explicitly specify one. */
	public static final Font.Style DEFAULT_STYLE = Font.Style.PLAIN;
	/** The default size for fonts which don't explicitly specify one. */
	public static final int DEFAULT_SIZE = 12;
	/** The default spacing for fonts which don't explicitly specify one. */
	public static final float DEFAULT_SPACING = 0.0f;
	
	/** The family name of this font. */
	private String _familyName;
	/** The style of this font (bold/italic). */
	private Style _style;
	/** The size of this font. */
	private int _size;
	/** The spacing of this font. */
	private float _spacing;
	
	/** Standard constructor. */
	public Font(String familyName, Style style, int size, float spacing)
	{
		if (familyName == null) familyName = DEFAULT_FAMILY_NAME;
		if (style == null) style = DEFAULT_STYLE;
		
		_familyName = familyName;
		_style = style;
		_size = size;
		_spacing = spacing;
	}
	
	/** Convenience constructor. */
	public Font(String familyName, Style style, int size)
	{
		this(familyName, style, size, DEFAULT_SPACING);
	}

	/** Convenience constructor. */
	public Font(String familyName, Style style)
	{
		this(familyName, style, DEFAULT_SIZE);
	}

	/** Convenience constructor. */
	public Font(String familyName)
	{
		this(familyName, DEFAULT_STYLE);
	}

	/** Convenience constructor. */
	public Font()
	{
		this(DEFAULT_FAMILY_NAME);
	}
	
	/** Copy constructor. */
	public Font(Font other)
	{
		_familyName = other._familyName;
		_style = other._style;
		_size = other._size;
		_spacing = other._spacing;
	}
	
	/** Gets the family name of this font. */
	public String getFamilyName()
	{
		return _familyName;
	}
	
	/** Sets the style of this font. */
	public Font setStyle(Style style)
	{
		_style = style;
		
		return this;
	}

	/** Gets the style of this font. */
	public Style getStyle()
	{
		return _style;
	}
	
	/** Returns true if this font has bold styling. */
	public boolean isBold()
	{
		return _style == Style.BOLD || _style == Style.BOLD_ITALIC;
	}

	/** Returns true if this font has italic styling. */
	public boolean isItalic()
	{
		return _style == Style.ITALIC || _style == Style.BOLD_ITALIC;
	}

	/** Returns true if this font has no bold/italic styling. */
	public boolean isPlain()
	{
		return _style == Style.PLAIN;
	}
	
	/** Removes bold styling if this font has it, or adds if if not. */
	public void toggleBold()
	{
		switch (_style) {
		case PLAIN:
			_style = Style.BOLD;
			break;
		case BOLD:
			_style = Style.PLAIN;
			break;
		case ITALIC:
			_style = Style.BOLD_ITALIC;
			break;
		case BOLD_ITALIC:
			_style = Style.ITALIC;
			break;
		}
	}

	/** Removes italic styling if this font has it, or adds if if not. */
	public void toggleItalic()
	{
		switch (_style) {
		case PLAIN:
			_style = Style.ITALIC;
			break;
		case BOLD:
			_style = Style.BOLD_ITALIC;
			break;
		case ITALIC:
			_style = Style.PLAIN;
			break;
		case BOLD_ITALIC:
			_style = Style.BOLD;
			break;
		}
	}
	
	/** Sets the size of this font. */
	public void setSize(int size)
	{
		_size = size;
	}

	/** Gets the size of this font. */
	public int getSize()
	{
		return _size;
	}

	/** Sets the spacing of this font. */
	public void setSpacing(float spacing)
	{
		_spacing = spacing;
	}

	/** Gets the spacing of this font. */
	public float getSpacing()
	{
		return _spacing;
	}
	
	@Override
	public Font clone()
	{
		return new Font(this);
	}

	// Automatically generated.
	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((_familyName == null) ? 0 : _familyName.hashCode());
		result = prime * result + _size;
		result = prime * result + Float.floatToIntBits(_spacing);
		result = prime * result + ((_style == null) ? 0 : _style.hashCode());
		return result;
	}

	// Automatically generated.
	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Font other = (Font) obj;
		if (_familyName == null) {
			if (other._familyName != null)
				return false;
		} else if (!_familyName.equals(other._familyName))
			return false;
		if (_size != other._size)
			return false;
		if (Float.floatToIntBits(_spacing) != Float.floatToIntBits(other._spacing))
			return false;
		if (_style != other._style)
			return false;
		return true;
	}
}
