package org.expeditee.gio;

import java.io.File;

import org.expeditee.core.Font;

/**
 * Class for managing font references in Expeditee. Expeditee's Font class is an
 * opaque handle that is translated to an internal representation when used by
 * other managers. The internal representation used for a given Font handle is
 * implementation-specific (i.e. if an exact match for the font handle is not registered
 * with the manager, the alternative selected is undefined).
 * 
 * TTF font files can be registered with the manager to increase the range of internal
 * fonts available, and a collection of handles to all of the fonts the manager supports
 * can be requested.
 * 
 * @author cts16
 */
public abstract class FontManager {

	/** The default font registered with the manager. */
	private Font _defaultFont = new Font();

	/** Returns the default font registered with the manager. */
	public final Font getDefaultFont()
	{
		return _defaultFont;
	}

	/** Sets the default font for the font manager. */
	public final void setDefaultFont(final Font font)
	{
		_defaultFont = font;
	}
	
	/** Registers a font file with the font manager, and returns the handle
	    to the newly registered font. */
	public abstract Font registerFontFile(File fontFile);
	
	/** Gets an array of all font families registered with the manager. */
	public abstract String[] getRegisteredFontFamilies();
	
	/** Returns true if the given font is registered with the manager. */
	public abstract boolean isFontFamilyRegistered(String fontFamily);

	/** Returns a handle representing the actual font used for the given font. */
	public abstract Font getActualFont(Font font);
	
	/** Checks if the given font was matched exactly by the font manager. */
	public boolean fontIsActualFont(Font font)
	{
		if (font == null) return false;
		
		return font.equals(getActualFont(font));
	}
	
	
}
