package org.expeditee.gio.gesture;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import org.expeditee.gio.gesture.data.GestureData;

/**
 * Class representing the abstract notion of input to Expeditee. Classes can implement
 * the <code>GestureListener</code> interface and register themselves with the <code>InputManager</code>
 * to receive notification of gestures, which are translated from the raw input (mouse/keyboard/touch/etc.).
 * 
 * @author cts16
 */
public class Gesture {
	
	/**
	 * Class representing a type of gesture. Identified by a string name. Each name corresponds
	 * to a single GestureType object throughout the entire program, so type matching can be
	 * performed with the == operator. The GestureType object also is initialised with a 
	 * GestureData type, which is the type of data that should come with gestures
	 * of that type.
	 * 
	 * @author cts16
	 */
	public static class GestureType
	{
		/** Static mapping retains reference to the single GestureType object per string. */
		private static HashMap<String, GestureType> _registeredTypes = new HashMap<String, GestureType>();
		
		/** The name of this gesture type. */
		private String _name;
		/** The type of data that goes with gestures of this type. */
		private Class<? extends GestureData> _dataClass;
		
		public static GestureType register(String name)
		{
			if (name == null || isRegistered(name)) return null;
			
			_registeredTypes.put(name, new GestureType(name, null));
			
			return get(name);
		}
		
		public static boolean isRegistered(String name)
		{
			return (name != null) && (_registeredTypes.containsKey(name));
		}
		
		public static GestureType get(String name)
		{
			if (name == null || !isRegistered(name)) return null;
			
			return _registeredTypes.get(name);
		}
		
		private GestureType(String name, Class<? extends GestureData> dataClass)
		{
			_name = name;
			_dataClass = dataClass;
		}
		
		public String name()
		{
			return _name;
		}
		
		public Class<? extends GestureData> dataClass()
		{
			return _dataClass;
		}
	}
	
	/** The type of gesture that was performed. */
	private GestureType _type;
	
	/** The specifics of the gesture (differs by gesture type). */
	private GestureData _data;
	
	/** Whether or not the gesture was created by the program instead of an actual input event. */
	private boolean _isRobotic = false;
	
	public Gesture(GestureType type, GestureData data)
	{
		assert(type != null);
		
		_type = type;
		_data = data;
	}
	
	public Gesture(Gesture other)
	{
		this(other._type, other._data);
		setRobotic(other._isRobotic);
	}
	
	public GestureType getType()
	{
		return _type;
	}
	
	public GestureData getData()
	{
		return _data;
	}
	
	/**
	 * @return
	 * 		True if the gesture was generated by software, false if it came from
	 * 		an actual input event.
	 */
	public boolean isRobotic()
	{
		return _isRobotic;
	}
	
	public void setRobotic(boolean isRobotic)
	{
		_isRobotic = isRobotic;
	}
	
	public Gesture clone()
	{
		return new Gesture(this);
	}
	
	/**
	 * Helper-method to create a list of gestures with a single gesture in it.
	 */
	public static List<Gesture> single(GestureType type, GestureData data)
	{
		return single(new Gesture(type, data));
	}
	
	/**
	 * Helper-method to create a list of gestures with a single gesture in it.
	 */
	public static List<Gesture> single(Gesture gesture)
	{
		List<Gesture> gestures = new LinkedList<Gesture>();
		gestures.add(gesture);
		return gestures;
	}
}
