package org.expeditee.gio.javafx;

import java.awt.AWTException;
import java.awt.Robot;

import org.expeditee.core.Dimension;
import org.expeditee.core.Line;
import org.expeditee.core.Point;
import org.expeditee.gio.EcosystemManager;
import org.expeditee.gio.InputManager;
import org.expeditee.gio.input.InputEvent;
import org.expeditee.gio.input.KBMInputEvent;
import org.expeditee.gio.input.InputEvent.InputType;
import org.expeditee.gio.javafx.JavaFXGraphicsManager.ExpediteeApplication;
import org.expeditee.items.Text;

import javafx.animation.KeyFrame;
import javafx.animation.Timeline;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.scene.input.KeyEvent;
import javafx.scene.input.MouseEvent;
import javafx.scene.input.ScrollEvent;
import javafx.stage.WindowEvent;
import javafx.util.Duration;

public class JavaFXInputManager extends InputManager {
	
	/** Singleton instance. */
	private static JavaFXInputManager _instance;
	
	/** Singleton instantiator. */
	public static JavaFXInputManager getInstance()
	{
		if (_instance == null) _instance = new JavaFXInputManager();
		
		return _instance;
	}
	
	// TODO: Is there a JavaFX way of controlling the mouse? cts16
	/** For robotically controlling the mouse position. */
	private Robot _robot;
	
	/** JavaFX sends spurious enter/exit events when pressing the middle mouse button. Ignore these. */
	boolean _lastExitSpurious = false;

	private JavaFXInputManager()
	{
		JavaFXInputManager manager = this;
		
		// Exit the program on window closed
		JavaFXGraphicsManager.ExpediteeApplication.theStage.setOnCloseRequest(new EventHandler<WindowEvent>() {
			@Override
			public void handle(WindowEvent event)
			{
				distributeWindowEvent(WindowEventType.WINDOW_CLOSED);
			}
		});
		
		// Register to receive window resize events
		ChangeListener<Number> windowSizeListener = new ChangeListener<Number>() {
			@Override
			public void changed(ObservableValue<? extends Number> observable, Number oldValue, Number newValue)
			{
				Dimension windowSize = EcosystemManager.getGraphicsManager().getWindowSize();
				ExpediteeApplication.theCanvas.setWidth(windowSize.width);
				ExpediteeApplication.theCanvas.setHeight(windowSize.height);
				distributeWindowEvent(WindowEventType.WINDOW_RESIZED);
			}
		};
		JavaFXGraphicsManager.ExpediteeApplication.theStage.widthProperty().addListener(windowSizeListener);
		JavaFXGraphicsManager.ExpediteeApplication.theStage.heightProperty().addListener(windowSizeListener);
		
		// Handle mouse wheel events
		JavaFXGraphicsManager.ExpediteeApplication.theScene.addEventHandler(ScrollEvent.SCROLL, new EventHandler<ScrollEvent>() {
			@Override
			public void handle(ScrollEvent event) {
				try {
					// Create an input event
					Integer rotation = new Integer((int) (event.getDeltaY() / event.getMultiplierY()));
					InputEvent expEvent = new KBMInputEvent(KBMInputEvent.EventType.MOUSE_WHEEL_SCROLL, rotation);
					
					// Translate and perform the input
					distributeInputEvent(expEvent);
				} catch (Exception exception) {
					System.err.println(exception.getMessage());
				}
			}
		});
		
		// Handle mouse movement events
		EventHandler<MouseEvent> mouseMovementEventHandler = new EventHandler<MouseEvent>() {
			@Override
			public void handle(MouseEvent event) {
				try {
					// Create an input event
					Point movedFrom = getCursorPosition();
					Point movedTo = new Point((int) event.getSceneX(), (int) event.getSceneY());
					InputEvent expEvent = new KBMInputEvent(KBMInputEvent.EventType.MOUSE_MOVE, new Line(movedFrom, movedTo));
					
					// Translate and perform the input
					distributeInputEvent(expEvent);
				} catch (Exception exception) {
					System.err.println(exception.getMessage());
				}
				
				// Update the cursor position
				updateCursorPosition((int) event.getSceneX(), (int) event.getSceneY());
			}
		};
		JavaFXGraphicsManager.ExpediteeApplication.theScene.addEventHandler(MouseEvent.MOUSE_DRAGGED, mouseMovementEventHandler);
		JavaFXGraphicsManager.ExpediteeApplication.theScene.addEventHandler(MouseEvent.MOUSE_MOVED, mouseMovementEventHandler);
		
		// Handle mouse press/release events
		EventHandler<MouseEvent> mousePressReleaseEventHandler = new EventHandler<MouseEvent>() {
			@Override
			public void handle(MouseEvent event) {
				// Work out which button was pressed
				KBMInputEvent.MouseButton button;
				switch (event.getButton()) {
					case PRIMARY:
						button = KBMInputEvent.MouseButton.LEFT;
						break;
					case MIDDLE:
						button = KBMInputEvent.MouseButton.MIDDLE;
						break;
					case SECONDARY:
						button = KBMInputEvent.MouseButton.RIGHT;
						break;
					default:
						return;
				}
				
				try {
					// Create an input event
					KBMInputEvent.EventType type = (event.getEventType() == MouseEvent.MOUSE_PRESSED) ? KBMInputEvent.EventType.MOUSE_BUTTON_DOWN : KBMInputEvent.EventType.MOUSE_BUTTON_UP;
					InputEvent expEvent = new KBMInputEvent(type, button);
					
					// Translate and perform the input
					distributeInputEvent(expEvent);
				} catch (Exception exception) {
					System.err.println(exception.getMessage());
				}
			}
		};
		JavaFXGraphicsManager.ExpediteeApplication.theScene.addEventHandler(MouseEvent.MOUSE_PRESSED, mousePressReleaseEventHandler);
		JavaFXGraphicsManager.ExpediteeApplication.theScene.addEventHandler(MouseEvent.MOUSE_RELEASED, mousePressReleaseEventHandler);

		
		// Handle mouse-enters-window events
		JavaFXGraphicsManager.ExpediteeApplication.theCanvas.addEventHandler(MouseEvent.MOUSE_ENTERED, new EventHandler<MouseEvent>() {
			@Override
			public void handle(MouseEvent event) {
				if (!_lastExitSpurious) distributeWindowEvent(WindowEventType.MOUSE_ENTERED_WINDOW);
			}
		});
		
		// Handle mouse-exits-window events
		JavaFXGraphicsManager.ExpediteeApplication.theCanvas.addEventHandler(MouseEvent.MOUSE_EXITED, new EventHandler<MouseEvent>() {
			@Override
			public void handle(MouseEvent event) {
				_lastExitSpurious = event.isMiddleButtonDown();
				if (!_lastExitSpurious) distributeWindowEvent(WindowEventType.MOUSE_EXITED_WINDOW);
			}
		});
		
		// Handle key-typed events
		JavaFXGraphicsManager.ExpediteeApplication.theScene.addEventHandler(KeyEvent.KEY_TYPED, new EventHandler<KeyEvent>() {
			@Override
			public void handle(KeyEvent event) {
				// Ignore control characters
				if (event.isControlDown() || event.isAltDown()) return;
				
				String typedText = event.getCharacter();
				
				for (int i = 0; i < typedText.length(); i++) {
					try {
						// Ignore Esc typing
						if (typedText.charAt(i) == Text.ESC_CHARACTER) continue;
						
						// Create an input event
						Character character = new Character(typedText.charAt(i));
						InputEvent expEvent = new KBMInputEvent(KBMInputEvent.EventType.CHAR_TYPED, character);
						
						// Translate and perform the input
						distributeInputEvent(expEvent);
					} catch (Exception exception) {
						System.err.println(exception.getMessage());
					}
				}
			}
		});
		
		// Handle key pressed/released events
		EventHandler<KeyEvent> keyPressReleaseEventHandler = new EventHandler<KeyEvent>() {
			@Override
			public void handle(KeyEvent event) {

				KBMInputEvent.Key key = JavaFXConversions.fromJavaFXKeyCode(event.getCode());
				if (key == null) return;
				
				try {
					// Create an input event
					KBMInputEvent.EventType type = (event.getEventType() == KeyEvent.KEY_PRESSED) ? KBMInputEvent.EventType.KEY_DOWN : KBMInputEvent.EventType.KEY_UP;
					InputEvent expEvent = new KBMInputEvent(type, key);
					
					// Translate and perform the input
					distributeInputEvent(expEvent);
				} catch (Exception exception) {
					System.err.println(exception.getMessage());
				}
			}
		};
		JavaFXGraphicsManager.ExpediteeApplication.theScene.addEventHandler(KeyEvent.KEY_PRESSED, keyPressReleaseEventHandler);
		JavaFXGraphicsManager.ExpediteeApplication.theScene.addEventHandler(KeyEvent.KEY_RELEASED, keyPressReleaseEventHandler);
		
		// Set up the robot (for controlling the mouse)
		// TODO: What to do if robot throws exception??? cts16
		try {
			_robot = new Robot();
		} catch (AWTException e) {
			e.printStackTrace();
		}
		
		_timer = new Timeline();
		_timer.cycleCountProperty().set(1);
	}

	@Override
	protected Point getRealCursorPosition()
	{
		// Can't get the real mouse position. Not a big deal as only used for initialisation
		// and is forgotten as soon as the mouse moves.
		return new Point(0, 0);
	}
	
	@Override
	public void setCursorPosition(Point position)
	{
		Point screenPosition = windowToScreenPosition(position);
		_robot.mouseMove(screenPosition.x, screenPosition.y);
	}

	@Override
	protected boolean isInputTypeSupported(InputType type)
	{
		// Only keyboard/mouse input is currently supported
		if (type == InputEvent.InputType.KBM) return true;
		
		return false;
	}
	
	/** Converts a location given in window-space to screen-space. */
	private Point windowToScreenPosition(Point windowPosition)
	{
		Point windowTopLeft = EcosystemManager.getGraphicsManager().getWindowLocation();
		
		return windowPosition.clone().add(windowTopLeft);
	}

	private Timeline _timer;
	
	@Override
	protected void updateTimer(long nextTimeout)
	{
		JavaFXInputManager jfxInputManager = this;
		
		if (nextTimeout < 0) nextTimeout = 0;
		_timer.stop();
		_timer.getKeyFrames().clear();
		_timer.getKeyFrames().add(new KeyFrame(new Duration(nextTimeout), new EventHandler<ActionEvent>() {
			@Override
			public void handle(ActionEvent event) {
				jfxInputManager.triggerTimeoutEvents();
			}
		}));
		_timer.playFromStart();
	}

}
