package org.expeditee.gio.javafx;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import org.expeditee.core.Font;
import org.expeditee.core.Line;
import org.expeditee.core.Point;
import org.expeditee.core.TextHitInfo;
import org.expeditee.core.TextLayout;
import org.expeditee.core.bounds.AxisAlignedBoxBounds;
import org.expeditee.gio.TextLayoutManager;

import javafx.geometry.Bounds;
import javafx.scene.text.Text;
import javafx.scene.text.TextAlignment;

public class JavaFXTextLayoutManager extends TextLayoutManager {
	
	/** Singleton instance. */
	private static JavaFXTextLayoutManager _instance;
	
	/** Singleton instantiator. */
	public static JavaFXTextLayoutManager getInstance()
	{
		if (_instance == null) _instance = new JavaFXTextLayoutManager();
		
		return _instance;
	}
	
	/** Mapping from Expeditee text-layout objects to JavaFX equivalents. */
	private HashMap<Long, javafx.scene.text.Text> _layoutMap;
	
	private JavaFXTextLayoutManager()
	{
		_layoutMap = new HashMap<Long, javafx.scene.text.Text>();
	}
	
	private synchronized void register(TextLayout layout, javafx.scene.text.Text jfxLayout)
	{
		if (layout == null || jfxLayout == null) return;
		
		_layoutMap.put(layout.getHandle(), jfxLayout);
	}
	
	public synchronized javafx.scene.text.Text getInternalLayout(TextLayout layout)
	{
		if (!isTextLayoutValid(layout)) return null;
		
		return _layoutMap.get(layout.getHandle());
	}
	
	public synchronized boolean isTextLayoutValid(TextLayout layout)
	{
		if (layout == null) return false;
		
		return _layoutMap.containsKey(layout.getHandle());
	}

	@Override
	public float getAdvance(TextLayout layout)
	{
		if (!isTextLayoutValid(layout)) return Float.NaN;
		
		Text text = getInternalLayout(layout);
		
		Bounds textBounds = text.getBoundsInLocal();
		
		return (float) textBounds.getWidth();
	}

	@Override
	public int getCharacterCount(TextLayout layout)
	{
		if (!isTextLayoutValid(layout)) return 0;
	
		Text text = getInternalLayout(layout);
		
		return text.getText().length();
	}

	@Override
	public TextHitInfo getNextLeftHit(TextLayout layout, int offset)
	{
		// TODO: Complete. cts16
		return new TextHitInfo(offset, false);
	}

	@Override
	public TextHitInfo getNextLeftHit(TextLayout layout, TextHitInfo hit)
	{
		// TODO: Complete. cts16
		return hit;
	}

	@Override
	public TextHitInfo getNextRightHit(TextLayout layout, int offset)
	{
		// TODO: Complete. cts16
		return new TextHitInfo(offset, false);
	}

	@Override
	public TextHitInfo getNextRightHit(TextLayout layout, TextHitInfo hit)
	{
		// TODO: Complete. cts16
		return hit;
	}

	@Override
	public float[] getCaretInfo(TextLayout layout, TextHitInfo hit)
	{
		// TODO: Complete. cts16
		float[] ret = new float[2];
		ret[0] = 0;
		ret[1] = 1;
		return ret;
	}

	@Override
	public TextHitInfo hitTestChar(TextLayout layout, float x, float y)
	{
		// TODO: Complete. cts16
		return new TextHitInfo(0, false);
	}

	@Override
	public AxisAlignedBoxBounds getPixelBounds(TextLayout layout, float x, float y)
	{
		if (!isTextLayoutValid(layout)) return null;
		
		Text text = getInternalLayout(layout);
		
		Bounds textBounds = text.getBoundsInLocal();
		
		return new AxisAlignedBoxBounds((int) x, (int) y, (int) textBounds.getWidth(), (int) textBounds.getHeight());
	}

	@Override
	public AxisAlignedBoxBounds getLogicalHighlightShape(TextLayout layout, int firstEndpoint, int secondEndpoint)
	{
		// TODO: Complete. cts16
		return getPixelBounds(layout, 0, -layout.getAscent());
	}

	@Override
	public float getAscent(TextLayout layout)
	{
		if (!isTextLayoutValid(layout)) return Float.NaN;
		
		Text text = getInternalLayout(layout);
		
		return (int) text.getBaselineOffset();
	}

	@Override
	public float getDescent(TextLayout layout)
	{
		if (!isTextLayoutValid(layout)) return Float.NaN;
		
		Text text = getInternalLayout(layout);
		
		return (int) (text.getBoundsInLocal().getHeight() - text.getBaselineOffset());
	}

	@Override
	public float getLeading(TextLayout layout)
	{
		// TODO Auto-generated method stub
		return 0;
	}

	// TODO: Remove magic constant of -1. cts16
	@Override
	public int getStringWidth(Font font, String string)
	{
		if (font == null || string == null) return -1;

		// Make sure we have a JavaFX font to use
		javafx.scene.text.Font jfxFont;
		JavaFXFontManager fontManager = JavaFXMiscManager.getIfUsingJavaFXFontManager();
		if (fontManager == null) return -1;
		jfxFont = fontManager.getInternalFont(font);
		if (jfxFont == null) {
			JavaFXGraphicsManager g = JavaFXMiscManager.getIfUsingJavaFXGraphicsManager();
			if (g == null) return -1;
			jfxFont = fontManager.getInternalFont(fontManager.getDefaultFont());
		}
		if (jfxFont == null) return -1;
		
		// Create a text object to layout the text
		javafx.scene.text.Text text = new javafx.scene.text.Text(string);
		text.setFont(jfxFont);
		
		// Return the width of the text
		return (int) text.getBoundsInLocal().getWidth();
	}

	@Override
	public void releaseLayout(TextLayout layout)
	{
		if (!isTextLayoutValid(layout)) return;
		
		_layoutMap.remove(layout.getHandle());
	}

	@Override
	public List<TextLayout> layoutString(String string, Font font, Point start, Line[] lineBreakers, int widthLimit,
			int lineSpacing, boolean dontBreakWords, boolean fullJustify)
	{
		if (string == null || font == null || start == null) return null;
		
		// Make sure we have a JavaFX font to use
		javafx.scene.text.Font jfxFont;
		JavaFXFontManager fontManager = JavaFXMiscManager.getIfUsingJavaFXFontManager();
		if (fontManager == null) return null;
		jfxFont = fontManager.getInternalFont(font);
		if (jfxFont == null) return null;
		
		// Temporary list to accumulate the Texts in to
		List<javafx.scene.text.Text> layouts = new LinkedList<javafx.scene.text.Text>();
		List<Integer> offsets = new LinkedList<Integer>();
		offsets.add(0);
		List<Point> positions = new LinkedList<Point>();
		positions.add(new Point(start));
		
		int startIndex = 0;
		while (startIndex < string.length()) {
			
			int width = widthLimit;
			if (lineBreakers != null) width = Math.min(width, (int) getLineWidth(start, lineBreakers));
			
			int endIndex = string.length();
			for (int i = startIndex + 1; i < string.length(); i++) {
				if (string.charAt(i) == '\n') {
					endIndex = i;
					break;
				}
			}
			
			javafx.scene.text.Text text = findTextNoWiderThan(string.substring(startIndex, endIndex), jfxFont, width, dontBreakWords);
			
			// If it's impossible to layout any more text without breaking a word, just do it
			if (text == null || text.getText().length() == 0 && width == widthLimit) {
				text = findTextNoWiderThan(string.substring(startIndex, endIndex), jfxFont, width, false);
				// If still impossible, give up
				if (text == null || text.getText().length() == 0) break;
			}
			
			endIndex = startIndex + text.getText().length();
			text.setX(start.x);
			text.setY(start.y);

			if (fullJustify && endIndex < string.length()) {
				text = getJustifiedLayout(text, width);
			}
			
			layouts.add(text);
			offsets.add(endIndex);
			startIndex = endIndex;
			if (startIndex < string.length() && string.charAt(startIndex) == '\n') startIndex += 1;
			int lineDrop = (int) (text.getBoundsInLocal().getHeight());
			if (lineSpacing >= 0) {
				lineDrop += lineSpacing; 
			}
			start.y += lineDrop;
			positions.add(new Point(start));
		}
		
		// Convert the accumulated list into an internal-type array
		List<TextLayout> ret = new LinkedList<TextLayout>();
		for (int i = 0; i < layouts.size(); i++) {
			javafx.scene.text.Text text = layouts.get(i);
			int startCharIndex = offsets.get(i);
			int endCharIndex = offsets.get(i + 1) - 1;
			String line = string.substring(startCharIndex, endCharIndex + 1);
			TextLayout layout = TextLayout.get(line, font, startCharIndex, endCharIndex);
			register(layout, text);
			ret.add(layout);
		}

		return ret;
	}
	
	private javafx.scene.text.Text findTextNoWiderThan(String string, javafx.scene.text.Font font, int maxWidth, boolean dontBreakWords)
	{
		if (string == null || font == null || maxWidth <= 0) return null;
		
		javafx.scene.text.Text text = new javafx.scene.text.Text(string);
		text.setFont(font);
		
		// TODO: Change to binary search. cts16
		while (text.getBoundsInLocal().getWidth() > maxWidth && string.length() > 0) {
			string = string.substring(0, string.length() - 1);
			if (dontBreakWords && !string.endsWith(" ")) continue;
			text.setText(string);
		}
		
		return text;
	}
	
	private javafx.scene.text.Text getJustifiedLayout(javafx.scene.text.Text text, int width)
	{
		if (text == null) return null;
		
		text.setTextAlignment(TextAlignment.JUSTIFY);
		text.setWrappingWidth(width);
		
		return text;
	}

}
